from django.http import HttpResponse
from django.views.generic import View
from django.conf import settings

from lino.utils.xmlgen import html as xghtml

class Index(View):

    def get(self, request, *args, **kwargs):
        ui = settings.LINO.groph_ui
        print repr(settings.LINO.modules.lino.Home.default_action.get_window_layout())
        actor = settings.LINO.modules.lino.Home
        ah = actor.get_handle(ui)
        print ah
        a = actor.default_action
        ar = actor.request(ui,request,a)
        print ar
        html = '\n'.join([ln for ln in self.html_page(ar,*args,**kwargs)])
        return HttpResponse(html)

    def html_page(self,ar,*args,**kw):
        wl = ar.action.get_window_layout()
        print wl.main
        lh = wl.get_layout_handle(ar.ui)
        print lh.main.__html__(ar)
        
        yield '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">'
        yield '<html><head>'
        yield '<meta http-equiv="Content-Type" content="text/html; charset=utf-8" />'
        yield '<title id="title">%s</title>' % settings.LINO.title
        
        user = ar.get_user()  # request.subst_user or request.user
        menu = settings.LINO.get_site_menu(self,user.profile)
        yield '</head><body>'
        
        yield menu.as_html(ar)
        story = []
        #~ story.append(xghtml.E.p(str(menu)))
        #~ story.append(xghtml.E.p(str(args)))
        #~ story.append(xghtml.E.p(str(kw)))
        e = xghtml.E.div(*story,id='body')
        yield xghtml.E.tostring(e)
        yield lh.main.__html__(ar)
        
        yield "</body></html>"
        



## Copyright 2009-2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

"Documented in :doc:`/topics/dumpy`."

import logging
logger = logging.getLogger(__name__)


from StringIO import StringIO
import os
import imp
from decimal import Decimal
#~ from types import GeneratorType


from django.conf import settings
from django.db import models
from django.db import IntegrityError
from django.db.models.fields import NOT_PROVIDED
from django.core.serializers import base
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.contrib.contenttypes.models import ContentType
#~ from django.contrib.auth.models import Permission
from django.contrib.sessions.models import Session
from django.utils.encoding import smart_unicode, is_protected_type, force_unicode

import lino
from lino.core.modeltools import obj2str, sorted_models_list, full_model_name
from lino.utils import dblogger
from lino.utils import babel
from lino import dd
#~ from lino.utils.mti import MtiChildWrapper

SUFFIX = '.py'
#~ SUFFIX = '.dpy'

class Serializer(base.Serializer):
    """
    Serializes a QuerySet to a py stream.
    Usage: ``manage.py dumpdata --format py``
    """

    internal_use_only = False
    
    write_preamble = True # may be set to False e.g. by testcases
    models = None
    
    def serialize(self, queryset, **options):
        self.options = options

        self.stream = options.get("stream", StringIO())
        self.selected_fields = options.get("fields")
        self.use_natural_keys = options.get("use_natural_keys", False)
        if self.write_preamble:
            self.stream.write('# -*- coding: UTF-8 -*-\n')
            #~ self.stream.write('# Created using Lino version %s\n' % lino.__version__)
            name,current_version,url = settings.LINO.using().next()
            if '+' in current_version:
                raise Exception(
                    "Cannot dumpdata from intermediate version %s" % current_version)
            
            self.stream.write('''\
"""
This is a `Python dump <http://lino-framework.org/topics/dumpy.html>`_ 
created by `%s <%s>`_.
"""
''' % (name,url))
            self.stream.write('SOURCE_VERSION = %r\n' % current_version)
            #~ self.stream.write('SOURCE_VERSION = %r\n' % lino.__version__)
            self.stream.write('from decimal import Decimal\n')
            self.stream.write('from datetime import datetime as dt\n')
            self.stream.write('from datetime import time,date\n')
            #~ self.stream.write('from lino.utils import i2d\n')
            #~ self.stream.write('from lino.utils.mti import insert_child\n')
            self.stream.write('from lino.utils.mti import create_child\n')
            self.stream.write('from lino.dd import resolve_model\n')
            self.stream.write('from django.contrib.contenttypes.models import ContentType\n')
            self.stream.write('from django.conf import settings\n')
            self.stream.write('''
            
def new_content_type_id(m):
    if m is None: return m
    # if not fmn: return None
    # m = resolve_model(fmn)
    ct = ContentType.objects.get_for_model(m)
    if ct is None: return None
    return ct.pk
                
''')
        #~ model = queryset.model
        if self.models is None:
            self.models = sorted_models_list() # models.get_models()
        if self.write_preamble:
            for model in self.models:
                self.stream.write('%s = resolve_model("%s")\n' % (
                  full_model_name(model,'_'), full_model_name(model)))
        self.stream.write('\n')
        for model in self.models:
            fields = [f for f,m in model._meta.get_fields_with_model() if m is None]
            #~ fields = model._meta.local_fields
            #~ fields = [f for f in model._meta.fields if f.serialize]
            #~ fields = [f for f in model._meta.local_fields if f.serialize]
            self.stream.write('def create_%s(%s):\n' % (
                model._meta.db_table,', '.join([f.attname for f in fields])))
            if model._meta.parents:
                if len(model._meta.parents) != 1:
                    msg = "%s : model._meta.parents is %r" % (model,model._meta.parents)
                    raise Exception(msg)
                pm,pf = model._meta.parents.items()[0]
                child_fields = [f for f in fields if f != pf]
                if child_fields:
                    attrs = ','+','.join([
                      '%s=%s' % (f.attname,f.attname) 
                          for f in child_fields])
                else: attrs = ''
                #~ self.stream.write('    return insert_child(%s.objects.get(pk=%s),%s%s)\n' % (
                    #~ full_model_name(pm,'_'),pf.attname,full_model_name(model,'_'),attrs))
                self.stream.write('    return create_child(%s,%s,%s%s)\n' % (
                    full_model_name(pm,'_'),pf.attname,full_model_name(model,'_'),attrs))
            else:
                for f in fields:
                    if isinstance(f,models.DecimalField):
                        self.stream.write(
                            '    if %s is not None: %s = Decimal(%s)\n' % (
                            f.attname,f.attname,f.attname))
                    if isinstance(f,models.ForeignKey) and f.rel.to is ContentType:
                        #~ self.stream.write(
                            #~ '    %s = ContentType.objects.get_for_model(%s).pk\n' % (
                            #~ f.attname,f.attname))
                        self.stream.write(
                            '    %s = new_content_type_id(%s)\n' % (
                            f.attname,f.attname))
                self.stream.write('    return %s(%s)\n' % (
                    full_model_name(model,'_'),
                    ','.join([
                        '%s=%s' % (f.attname,f.attname) for f in fields])))
        #~ self.start_serialization()
        self.stream.write('\n')
        model = None
        all_models = []
        for obj in queryset:
            if isinstance(obj,ContentType): continue
            #~ if isinstance(obj,Session): continue
            #~ if isinstance(obj,Permission): continue
            if obj.__class__ != model:
                model = obj.__class__
                if model in all_models:
                    raise Exception("%s instances weren't grouped!" % model)
                all_models.append(model)
                self.stream.write('\ndef %s_objects():\n' % model._meta.db_table)
            fields = [f for f,m in model._meta.get_fields_with_model() if m is None]
            #~ fields = obj._meta.local_fields
            #~ fields = [f for f in obj._meta.local_fields if f.serialize]
            self.stream.write('    yield create_%s(%s)\n' % (
                obj._meta.db_table,
                ','.join([self.value2string(obj,f) for f in fields])))
        self.stream.write('\n\ndef objects():\n')
        all_models = self.sort_models(all_models)
        for model in all_models:
            #~ self.stream.write('    for o in %s_objects(): yield o\n' % model._meta.db_table)
            self.stream.write('    yield %s_objects()\n' % model._meta.db_table)
        self.stream.write('\nsettings.LINO.loading_from_dump = True\n')
        self.stream.write('\nsettings.LINO.install_migrations(globals())\n')
        #~ if settings.LINO.migration_module:
            #~ self.stream.write('\n')
            #~ self.stream.write('from %s import install\n' \
                #~ % settings.LINO.migration_module)
            #~ self.stream.write('install(globals())\n')
            
    def sort_models(self,unsorted):
        sorted = []
        hope = True
        """
        20121120 if we convert the list to a set, we gain some performance 
        for the ``in`` tests, but we obtain a random sorting order for all 
        independent models, making the double dump test less evident.
        """
        #~ 20121120 unsorted = set(unsorted)
        while len(unsorted) and hope:
            hope = False
            guilty = dict()
            #~ print "hope for", [m.__name__ for m in unsorted]
            for model in unsorted:
                deps = set([f.rel.to 
                  for f in model._meta.fields 
                      if f.rel is not None and f.rel.to is not model and f.rel.to in unsorted])
                #~ deps += [m for m in model._meta.parents.keys()]
                for m in sorted:
                    if m in deps:
                        deps.remove(m)
                if len(deps):
                    guilty[model] = deps
                else:
                    sorted.append(model)
                    unsorted.remove(model)
                    hope = True
                    break
                    
                #~ ok = True
                #~ for d in deps:
                    #~ if d in unsorted:
                        #~ ok = False
                #~ if ok:
                    #~ sorted.append(model)
                    #~ unsorted.remove(model)
                    #~ hope = True
                    #~ break
                #~ else:
                    #~ guilty[model] = deps
                #~ print model.__name__, "depends on", [m.__name__ for m in deps]
        if unsorted:
            assert len(unsorted) == len(guilty)
            msg = "There are %d models with circular dependencies :\n" % len(unsorted)
            msg += "- " + '\n- '.join([
                full_model_name(m)+' (depends on %s)' % ", ".join([full_model_name(d) for d in deps]) for m,deps in guilty.items()])
            for ln in msg.splitlines():
                self.stream.write('\n    # %s' % ln)
            dblogger.info(msg)
            sorted.extend(unsorted)              
        return sorted      
    

    #~ def start_serialization(self):
        #~ self._current = None
        #~ self.objects = []

    #~ def end_serialization(self):
        #~ pass

    #~ def start_object(self, obj):
        #~ self._current = {}

    #~ def end_object(self, obj):
        #~ self.objects.append({
            #~ "model"  : smart_unicode(obj._meta),
            #~ "pk"     : smart_unicode(obj._get_pk_val(), strings_only=True),
            #~ "fields" : self._current
        #~ })
        #~ self._current = None

    def value2string(self, obj, field):
        value = field._get_val_from_obj(obj)
        # Protected types (i.e., primitives like None, numbers, dates,
        # and Decimals) are passed through as is. All other values are
        # converted to string first.
        if value is None:
        #~ if value is None or value is NOT_PROVIDED:
            return 'None'
        if isinstance(field,models.DateTimeField):
            d = value
            return 'dt(%d,%d,%d,%d,%d,%d)' % (
              d.year,d.month,d.day,d.hour,d.minute,d.second)
        if isinstance(field,models.TimeField):
            d = value
            return 'time(%d,%d,%d)' % (d.hour,d.minute,d.second)
        if isinstance(field,models.ForeignKey) and field.rel.to is ContentType:
            ct = ContentType.objects.get(pk=value)
            return full_model_name(ct.model_class(),'_')
            #~ return "'"+full_model_name(ct.model_class())+"'"
            #~ return repr(tuple(value.app_label,value.model))
        if isinstance(field,models.DateField):
            d = value
            return 'date(%d,%d,%d)' % (d.year,d.month,d.day)
            #~ return 'i2d(%4d%02d%02d)' % (d.year,d.month,d.day)
        if isinstance(value,(float,Decimal)):
            return repr(str(value))
        if isinstance(value,(int,long)):
            return str(value)
        return repr(field.value_to_string(obj))

    def handle_fk_field(self, obj, field):
        related = getattr(obj, field.name)
        if related is not None:
            if self.use_natural_keys and hasattr(related, 'natural_key'):
                related = related.natural_key()
            else:
                if field.rel.field_name == related._meta.pk.name:
                    # Related to remote object via primary key
                    related = related._get_pk_val()
                else:
                    # Related to remote object via other field
                    related = smart_unicode(getattr(related, field.rel.field_name), strings_only=True)
        self._current[field.name] = related

    def handle_m2m_field(self, obj, field):
        if field.rel.through._meta.auto_created:
            if self.use_natural_keys and hasattr(field.rel.to, 'natural_key'):
                m2m_value = lambda value: value.natural_key()
            else:
                m2m_value = lambda value: smart_unicode(value._get_pk_val(), strings_only=True)
            self._current[field.name] = [m2m_value(related)
                               for related in getattr(obj, field.name).iterator()]

    #~ def getvalue(self):
        #~ return self.objects

SUPPORT_EMPTY_FIXTURES = False # trying, but doesn't yet work

if SUPPORT_EMPTY_FIXTURES:
    from lino.utils import AttrDict
    class DummyDeserializedObject(base.DeserializedObject):
        class FakeObject:
            _meta = AttrDict(db_table='')
        object = FakeObject()
        def __init__(self): pass
        def save(self, *args,**kw): pass

class FakeDeserializedObject(base.DeserializedObject):
    """
    Imitates DeserializedObject required by loaddata.
    
    Unlike normal DeserializedObject, we *don't want* to bypass 
    pre_save and validation methods on the individual objects.
    
    """
    
    
    def __init__(self, deserializer, object):
        self.object = object
        #~ self.name = name
        self.deserializer = deserializer

    def save(self, *args,**kw):
        """
        """
        #~ print 'dpy.py',self.object
        #~ dblogger.info("Loading %s...",self.name) 
        
        self.try_save(*args,**kw)
        #~ if self.try_save(*args,**kw):
            #~ self.deserializer.saved += 1
        #~ else:
            #~ self.deserializer.save_later.append(self)
        
    def try_save(self,*args,**kw):
        """Try to save the specified Model instance `obj`. Return `True` 
        on success, `False` if this instance wasn't saved and should be 
        deferred.
        """
        obj = self.object
        try:
            """
            """
            m = getattr(obj,'before_dumpy_save',None)
            if m is not None:
                m()
            obj.full_clean()
            obj.save(*args,**kw)
            dblogger.debug("%s has been saved" % obj2str(obj))
            self.deserializer.register_success()
            return True
        #~ except ValidationError,e:
        #~ except ObjectDoesNotExist,e:
        #~ except (ValidationError,ObjectDoesNotExist), e:
        #~ except (ValidationError,ObjectDoesNotExist,IntegrityError), e:
        except Exception, e:
            if not settings.LINO.loading_from_dump:
                # hand-written fixtures are expected to not raise any exception
                dblogger.warning("Failed to save %s:" % obj2str(obj))
                raise
            if False:
              """
              20120906 deactivated this test. also fixtures from dump may yield instances without pk.
              example migrate_from_1_4_10 adds pcsw.Third instances without pk, and which need to 
              get a second chance.
              """
              if obj.pk is None: 
                """
                (no longer true:) 
                presto.tim2lino creates invoices without pk which possibly fail to save 
                on first attempt because their project
                """
                if True:
                    msg = "Failed to save %s without %s: %s." % (
                        obj.__class__,obj._meta.pk.attname,obj2str(obj))
                    dblogger.warning(msg)
                    raise
                else:
                    dblogger.exception(e)
                    raise Exception(msg)
            deps = [f.rel.to for f in obj._meta.fields if f.rel is not None]
            if not deps:
                dblogger.exception(e)
                raise Exception("Failed to save independent %s." % obj2str(obj))
            self.deserializer.register_failure(self,e)
            return False
        #~ except Exception,e:
            #~ dblogger.exception(e)
            #~ raise Exception("Failed to save %s. Abandoned." % obj2str(obj))




#~ IS_DESERIALIZING = False

#~ def is_deserializing():
    #~ """
    #~ Some special features may want to know out whether they are happening 
    #~ during a deserialization or not.
    #~ The only one known so far is :func:`lino.modlib.cal.models.update_auto_task`, 
    #~ which should not do anything during deserialization.    
    #~ See :doc:`/blog/2011/0727`.
    #~ """
    #~ return IS_DESERIALIZING
    
class FlushDeferredObjects: 
    """
    Indicator class object. 
    Fixture may yield a ``dumpy.FlushDeferredObjects`` 
    to indicate that all deferred objects should get saved before going on.
    """
    pass

class DpyDeserializer:
    """
    Used when ``manage.py loaddata`` encounters a `.py` fixture.
    # new after blog/2011/0831
    """
    
    def __init__(self):
        #~ dblogger.info("20120225 DpyDeserializer.__init__()")
        self.save_later = {}
        self.saved = 0
        #~ self.count = 0

  
    def flush_deferred_objects(self):
        """
        Flush the list of deferred objects.
        """
        while self.saved and self.save_later:
            try_again = []
            for msg_objlist in self.save_later.values():
                for objlist in msg_objlist.values():
                    try_again += objlist
            logger.info("Trying again with %d unsaved instances.",
                len(try_again))
            self.save_later = {}
            self.saved = 0
            for obj in try_again:
                obj.try_save() # ,*args,**kw):
                #~ if obj.try_save(): # ,*args,**kw):
                    #~ self.saved += 1
                #~ else:
                    #~ self.save_later.append(obj)
            dblogger.info("Saved %d instances.",self.saved)
            
    def deserialize(self,fp, **options):
        #~ dblogger.info("20120225 DpyDeserializer.deserialize()")
        if isinstance(fp, basestring):
            raise NotImplementedError
        babel.set_language(babel.DEFAULT_LANGUAGE)
        
        #~ self.count += 1
        fqname = 'lino.utils.dumpy_tmp_%s' % hash(self)
        
        if False:
            parts = fp.name.split(os.sep)
            #~ parts = os.path.split(fp.name)
            print parts
            #~ fqname = parts[-1]
            fqname = '.'.join([p for p in parts if not ':' in p])
            assert fqname.endswith(SUFFIX)
            fqname = fqname[:-len(SUFFIX)]
            print fqname
        desc = (SUFFIX,'r',imp.PY_SOURCE)
        module = imp.load_module(fqname, fp, fp.name, desc)
        #~ module = __import__(filename)
        
        def expand(obj):
            if obj is None:
                pass # ignore None values
            elif obj is FlushDeferredObjects:
                self.flush_deferred_objects()
            elif isinstance(obj,models.Model):
                yield FakeDeserializedObject(self,obj)
            elif hasattr(obj,'__iter__'):
            #~ if type(obj) is GeneratorType:
                #~ dblogger.info("20120225 expand iterable %r",obj)
                for o in obj: 
                    for so in expand(o): 
                        yield so
            #~ elif isinstance(obj,MtiChildWrapper):
                # the return value of create_child()
                #~ yield FakeDeserializedObject(self,obj)
                #~ obj.deserializer = self
                #~ yield obj
            else:
                dblogger.warning("Ignored unknown object %r",obj)
                
        dblogger.info("Loading %s...",fp.name)
        if not hasattr(module,'objects'):
            raise Exception("%s has no attribute 'objects'" % fp.name)
        empty_fixture = True
        for obj in module.objects():
            for o in expand(obj): 
                empty_fixture = False
                yield o
        if empty_fixture:
            if SUPPORT_EMPTY_FIXTURES:
                yield DummyDeserializedObject() # avoid Django interpreting empty fixtures as an error
            else:
                """
                To avoid Django interpreting empty fixtures as an error, 
                we yield one object which always exists: the SiteConfig
                """
                yield FakeDeserializedObject(self,settings.LINO.site_config)
                #~ raise Exception("""\
#~ Fixture %s decided to not create any object.
#~ We're sorry, but Django doesn't like that. 
#~ See <https://code.djangoproject.com/ticket/18213>.
#~ """ % fp.name)
          
        #~ dblogger.info("Saved %d instances from %s.",self.saved,fp.name)
        
        self.flush_deferred_objects()
        
        if self.save_later:
            count = 0
            s = ''
            for model,msg_objects in self.save_later.items():
                for msg,objects in msg_objects.items():
                    if False: # detailed content of the first object
                        s += "\n- %s %s (%d object(s), e.g. %s)" % (
                          full_model_name(model),msg,len(objects),
                          obj2str(objects[0].object,force_detailed=True))
                    else: # pk of all objects
                        s += "\n- %s %s (%d object(s) with primary key %s)" % (
                          full_model_name(model),msg,len(objects),
                          ', '.join([unicode(o.object.pk) for o in objects]))
                    count += len(objects)
            
            msg = "Abandoning with %d unsaved instances from %s:%s" % (
                count,fp.name,s)
            dblogger.warning(msg)
            """
            Don't raise an exception. The unsaved instances got lost and 
            the loaddata should be done again, but meanwhile the database
            is not necessarily invalid and may be used for further testing.
            """
            #~ raise Exception(msg)
            
        if hasattr(module,'after_load'):
            module.after_load()
        #~ IS_DESERIALIZING = False

    def register_success(self):
        self.saved += 1
        
    def register_failure(self,obj,e):
        msg = force_unicode(e)
        d = self.save_later.setdefault(obj.object.__class__,{})
        l = d.setdefault(msg,[])
        if len(l) == 0:
            dblogger.info("Deferred %s : %s",obj2str(obj.object),msg)
        l.append(obj)

def Deserializer(fp, **options):
    d = DpyDeserializer()
    return d.deserialize(fp, **options)




from lino.ui import base
from lino.core import layouts

class Panel(object):
    def __init__(self,lh,name,vertical,*elems,**pkw):
        self.lh = lh
        self.name = name
        self.vertical = vertical
        self.elems = elems
        for k,v in pkw.items():
            setattr(self,k,v)
        #~ self.options = pkw
        
    def __html__(self,ar):
        if isinstance(self.lh.layout,layouts.ListLayout):
            return 'ext_elems.GridElement(lh,name,lh.layout._table,*elems,**pkw)'
        if isinstance(self.lh.layout,layouts.ParamsLayout) : 
            return 'ext_elems.ParamsPanel(lh,name,vertical,*elems,**pkw)'
        if isinstance(self.lh.layout,layouts.FormLayout): 
            if len(self.elems) == 1 or self.vertical:
                return 'ext_elems.DetailMainPanel(lh,name,vertical,*elems,**pkw)'
            else:
                return 'ext_elems.TabPanel(lh,name,*elems,**pkw)'
        raise Exception("No element class for layout %r" % self.lh.layout)
      
        
class UI(base.UI):
    """The central instance of Lino's ExtJS3 User Interface.
    """
    _handle_attr_name = '_groph_handle'
    #~ _response = None
    name = 'groph'
    verbose_name = "groph"
    
    def create_layout_element(self,lh,name,**kw):
        try:
            de = lh.get_data_elem(name)
            return de
        except Exception, e:
            de = None
            name += " (" + str(e) + ")"
            
        if hasattr(lh,'rh'):
            msg = "Unknown element %r referred in layout %s of %s." % (
                name,lh.layout,lh.rh.actor)
            l = [de.name for de in lh.rh.actor.wildcard_data_elems()]
            model = getattr(lh.rh.actor,'model',None) # VirtualTables don't have a model
            if getattr(model,'_lino_slaves',None):
                l += [str(rpt) for rpt in model._lino_slaves.values()]
            msg += " Possible names are %s." % ', '.join(l)
        else:
            msg = "Unknown element %r referred in layout %s." % (
                name,lh.layout)
            msg += "Cannot handle %r" % de
        raise KeyError(msg)


    def create_layout_panel(self,lh,name,vertical,elems,**kw):
        pkw = dict()
        pkw.update(labelAlign=kw.pop('label_align','top'))
        pkw.update(hideCheckBoxLabels=kw.pop('hideCheckBoxLabels',True))
        pkw.update(label=kw.pop('label',None))
        pkw.update(width=kw.pop('width',None))
        pkw.update(height=kw.pop('height',None))
        if kw:
            raise Exception("Unknown panel attributes %r" % kw)
        return Panel(lh,name,vertical,*elems,**pkw)





# -*- coding: utf-8 -*-
## Copyright 2011 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

"""

A setup funtion for Sphinx documentaton.

Used by the Lino documentation::

  from lino.utils.sphinx import setup

"""

import os
import lino


#~ from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned

from docutils import nodes
from docutils.parsers.rst import directives


#~ class ScreenshotDirective(directives.images.Image):
    #~ """
    #~ Directive to insert a screenshot.
    #~ """
    #~ def run(self):
        #~ assert len(self.arguments) == 1
        #~ # name = '/../gen/screenshots/' + self.arguments[0]
        #~ name = '/gen/screenshots/' + self.arguments[0]
        #~ self.arguments = [name]
        #~ (image_node,) = directives.images.Image.run(self)
        #~ return [image_node]





def srcref(mod):
    """
    Return the `source file name` for usage by Sphinx's ``srcref`` role.
    Returns None if the source file is empty (which happens e.g. for __init__.py 
    files whose only purpose is to mark a package).
    
    >>> from lino.utils import log
    >>> print srcref(log)
    lino/utils/log.py

    >>> from lino import utils
    >>> print srcref(utils)
    lino/utils/__init__.py
    
    >>> from lino.management import commands
    >>> print srcref(commands)
    None

    >>> from lino_welfare import settings
    >>> print srcref(settings)
    lino_welfare/settings.py

    """
    if not mod.__name__.startswith('lino.'): 
        return
    srcref = mod.__file__
    if srcref.endswith('.pyc'):
        srcref = srcref[:-1]
    if os.stat(srcref).st_size == 0:
        return 
    srcref = srcref[len(lino.__file__)-17:]
    srcref = srcref.replace(os.path.sep,'/')
    return srcref
    


def autodoc_skip_member(app, what, name, obj, skip, options):
    if name != '__builtins__':
        #~ print 'autodoc_skip_member', what, repr(name), repr(obj)
    
        if what == 'class':
            if name.endswith('MultipleObjectsReturned'):
                return True
            if name.endswith('DoesNotExist'):
                return True
                
            #~ if isinstance(obj,ObjectDoesNotExist) \
              #~ or isinstance(obj,MultipleObjectsReturned): 
                #~ return True
        
    #~ if what == 'exception': 
        #~ print 'autodoc_skip_member',what, repr(name), repr(obj), skip
        #~ return True
        

#~ SIDEBAR = """
#~ (This module's source code is available at :srcref:`/%s`)

#~ """  
#~ SIDEBAR = """
#~ (source code: :srcref:`/%s`)

#~ """  

SIDEBAR = """
(:srcref:`source code </%s>`)

"""  

#~ SIDEBAR = """
#~ .. sidebar:: Use the source, Luke

  #~ We generally recommend to also consult the source code.
  #~ This module's source code is available at
  #~ :srcref:`/%s.py`

#~ """  


    
def autodoc_add_srcref(app,what,name,obj,options,lines):
    if what == 'module':
        s = srcref(obj)
        if s:
            #~ srcref = name.replace('.','/')
            s = (SIDEBAR % s).splitlines()
            s.reverse()
            for ln in s:
                lines.insert(0,ln)
            #~ lines.insert(0,'')
            #~ lines.insert(0,'(We also recommend to read the source code at :srcref:`/%s.py`)' % name.replace('.','/'))
    

def setup(app):
    """
    The Sphinx setup function used for Lino-related documentation trees.
    """
    app.add_object_type(directivename='xfile',rolename='xfile',
      indextemplate='pair: %s; file')
    app.add_object_type(directivename='setting',rolename='setting',
      indextemplate='pair: %s; setting')
    #~ app.add_object_type(directivename='model',rolename='model',
      #~ indextemplate='pair: %s; model')
    #~ app.add_object_type(directivename='field',rolename='field',
      #~ indextemplate='pair: %s; field')
    app.add_object_type(directivename='table',rolename='table',
      indextemplate='pair: %s; table')
    app.add_object_type(directivename='screenshot',rolename='screen',
      indextemplate='pair: %s; screenshot')
    app.add_object_type(directivename='modattr',rolename='modattr',
      indextemplate='pair: %s; model attribute')
    app.add_object_type(directivename='model',rolename='model',
      indextemplate='pair: %s; model')
    #app.connect('build-finished', handle_finished)
    
    app.connect('autodoc-skip-member',autodoc_skip_member)
    app.connect('autodoc-process-docstring', autodoc_add_srcref)

    #~ app.add_directive('screenshot', ScreenshotDirective)
    #~ app.add_config_value('screenshots_root', '/screenshots/', 'html')

    app.add_stylesheet('linodocs.css')





# -*- coding: UTF-8 -*-
## Copyright 2009-2013 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.
"""
Adds an arbitrary selection of a few demo cities.
"""

from __future__ import unicode_literals

import logging
logger = logging.getLogger(__name__)

from django.core.exceptions import MultipleObjectsReturned
from lino.utils import dblogger
from lino.utils import babel
from lino.core.modeltools import resolve_model
from lino.utils.instantiator import Instantiator
from lino.utils.babel import babel_values


from lino import dd

def objects():
    #~ dblogger.info("Installing countries few_cities fixture")
    countries = dd.resolve_app('countries')
    #~ City = resolve_model('countries.City')
    City = countries.City
    Country = countries.Country
    CityTypes = countries.CityTypes
    city = Instantiator(City,'name country').build
    def make_city(name,country_id,**kw):
        #~ kw.update()
        flt = babel.lookup_filter('name',name,country__isocode=country_id,**kw)
        try:
            return City.objects.exclude(type=CityTypes.county).get(flt)
            #~ return City.objects.exclude(type=CityTypes.county).get(
                #~ country__isocode=country_id,name=name)
        except MultipleObjectsReturned:
            #~ qs = City.objects.exclude(type=CityTypes.county).filter(country__isocode=country_id,name=name)
            logger.info("Oops, there are multiple cities for %r", name)
            return qs[0]
        except City.DoesNotExist:
            return city(name,country_id,**kw)
        
    BE = Country.objects.get(pk='BE')
    DE = Country.objects.get(pk='DE')
    FR = Country.objects.get(pk='FR')
    yield make_city('Eupen','BE',zip_code='4700',type=CityTypes.city)
    yield City(country=BE,zip_code='4720',type=CityTypes.city,
      **babel_values('name',de='Kelmis',fr='La Calamine',en="Kelmis"))
    yield make_city('Kettenis','BE',zip_code='4701',type=CityTypes.village)
    yield make_city('Raeren','BE',zip_code='4730',type=CityTypes.village)
    yield make_city('Angleur','BE',zip_code='4031',type=CityTypes.city)
    yield City(country=BE,zip_code='4000',type=CityTypes.city,
      **babel_values('name',de='Lüttich',fr='Liège',en='Liège',nl="Luik"))
    yield City(country=BE,zip_code='1000',type=CityTypes.city,
      **babel_values('name',de='Brüssel',fr='Bruxelles',nl="Brussel",en="Brussels"))
    #~ yield city('Brussel','BE',zip_code='1000')
    #~ yield city(u'Brüssel','BE',zip_code='1000')
    yield City(country=BE,zip_code='8400',type=CityTypes.city,
      **babel_values('name',de='Ostende',fr='Ostende',nl="Oostende",en="Ostende"))
    
    harjumaa = make_city(u'Harjumaa','EE',type=CityTypes.county)
    yield harjumaa
    parnumaa = make_city(u'Pärnumaa','EE',type=CityTypes.county)
    yield parnumaa
    raplamaa = make_city(u'Raplamaa','EE',type=CityTypes.county)
    yield raplamaa
    
    yield make_city(u'Vigala','EE',type=CityTypes.municipality,parent=raplamaa)
    yield make_city(u'Rapla','EE',type=CityTypes.town,parent=raplamaa)
    
    yield make_city(u'Tallinn','EE',type=CityTypes.city,parent=harjumaa)
    yield make_city(u'Pärnu','EE',type=CityTypes.town,parent=parnumaa)
    yield make_city(u'Tartu','EE',type=CityTypes.town)
    yield make_city(u'Narva','EE',type=CityTypes.town)
    yield make_city(u'Ääsmäe','EE',type=CityTypes.town,parent=harjumaa)

    #~ yield make_city(u'Aachen','DE')
    yield City(country=DE,type=CityTypes.city,
      **babel_values('name',de='Aachen',fr='Aix-la-Chapelle',nl="Aken",en="Aachen"))
    yield City(country=DE,type=CityTypes.city,
      **babel_values('name',de='Köln',fr='Cologne',nl="Keulen",en="Cologne"))
    yield make_city(u'Berlin','DE')
    yield make_city(u'Hamburg','DE')
    yield City(country=DE,type=CityTypes.city,
      **babel_values('name',de='München',fr='Munich',en="Munich"))
    
    yield make_city(u'Maastricht','NL')
    yield make_city(u'Amsterdam','NL')
    yield make_city(u'Den Haag','NL')
    yield make_city(u'Rotterdam','NL')
    yield make_city(u'Utrecht','NL')
    yield make_city(u'Breda','NL')
    
    yield City(country=FR,type=CityTypes.city,
      **babel_values('name',de='Paris',fr='Paris',en="Paris",et="Pariis",nl="Parijs"))
    yield City(country=FR,type=CityTypes.city,
      **babel_values('name',de='Nizza',fr='Nice',en="Nice"))
    yield make_city(u'Metz','FR')
    yield make_city(u'Strasbourg','FR')
    yield make_city(u'Nancy','FR')
    yield make_city(u'Marseille','FR')



import os
from xml.dom import minidom 

def elements(node):
    e = node.firstChild
    while e is not None:
        if e.nodeType == minidom.Node.ELEMENT_NODE:
            yield e
        e = e.nextSibling

def xsd2xg(fn,nsname):
    dom = minidom.parse(fn) 
    yield "from lino.utils import xmlgen as xg"
    yield "class %s(xg.Namespace):" % nsname
    yield "  url = %r" % str(dom.documentElement.attributes.get('targetNamespace').value)
    #~ e = dom.documentElement.firstChild
    #~ while e is not None:
    indent = '  '
    for e in elements(dom.documentElement):
        if e.tagName == 'xsd:complexType':
            na = e.attributes.get('name',None)
            if na is None:
                yield indent + "# skipped nameless %s" % e.tagName
                continue
            yield indent + "# %s" % e.tagName
            yield indent + "class %s(xg.Container):" % na.value
            for ee in elements(e):
                #~ seq = e.getElementsByTagName('xsd:sequence')[0]
                if ee.tagName == 'xsd:sequence':
                    #~ seq = e.attributes.get('sequence',None)
                    #~ if seq:
                    for item in elements(ee):
                        na = item.attributes.get('name',None)
                        if na is None: continue
                        ta = item.attributes.get('type',None)
                        if ta is None:
                            yield indent*2 + "class %s(): pass" % na.value
                        else:
                            if ':' in ta.value:
                                typename = ta.value.replace(':','.')
                            else:
                                typename = nsname + '.' + ta.value
                            yield indent*2 + "class %s(%s): pass" % (na.value,typename)
        elif e.tagName == 'xsd:simpleType':
            na = e.attributes.get('name',None)
            if na is None:
                yield indent + "# skipped nameless %s" % e.tagName
                continue
            yield indent + "# %s" % e.tagName
            yield indent + "class %s(xg.Container):" % na.value
        else:
            yield indent + "# unhandled element %s" % e.tagName
    #~ e = e.nextSibling


fn = os.path.join('XSD','RetrieveTIGroupsV3.xsd')
#~ for i,ln in enumerate(text_lines(dom)):
for ln in xsd2xg(fn,'tx25'):
    #~ print "%d:%s" % (i,ln)
    print ln
    


# -*- coding: UTF-8 -*-
## Copyright 2011-2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

"""
Lino Così is yet another attempt to make Belgian accounting simple.

With Lino Così (when it is ready) you will be able to

- manage your contacts (organisations and persons)
- optionally manage a list of products (sales items)
- record your sales invoices (optionally printing them)
- manage your account chart 
- record your purchase invoices
- declare your VAT statement
- record your bank statements
- get your financial situation 

About the name: 
Le but est de proposer enfin un logiciel de comptabilité vraiment simple à utiliser.
Une **co**\ mptabilité **si**\ mple? 
"Così" en italien signifie "comme ceci".
Così! c'est comme ça qu'on aime la compta!

(Mais bien sûr il nous reste un peu de travail avant d'y être.
Bienvenus les volontaires pour tester 
http://demo4.lino-framework.org
et m'aider à réfléchir.)

"""

from os.path import join, abspath, dirname

from lino.apps.std.settings import *

from django.utils.translation import ugettext_lazy as _

class Lino(Lino):
  
    #~ title = __name__
    short_name = u"Lino Così"
    description = _("a Lino application to make Belgian accounting simple.")
    version = "0.1"
    url = "http://www.lino-framework.org/autodoc/lino.apps.cosi"
    author = 'Luc Saffre'
    author_email = 'luc.saffre@gmail.com'
    
    demo_fixtures = 'std few_countries few_cities few_languages demo demo2 userman'.split()
    
    languages = ['en','de','fr']
    #~ languages = 'de fr et en'.split()
    
    #~ project_model = 'tickets.Project'
    user_model = 'users.User'
    
    #~ remote_user_header = "REMOTE_USER"
    
    #~ override_modlib_models = [ 'contacts.Partner' ]
    #~ override_modlib_models = [
      #~ 'contacts.Person','contacts.Company',
      #~ 'households.Household']
    
    #~ def get_description(self):
        #~ from django.utils.translation import ugettext_lazy as _
        #~ from django.utils.translation import string_concat
        #~ return _("a Lino application to make Belgian accounting simple.")
    #~ description = property(get_description)
        
    def get_app_source_file(self): return __file__
      
    #~ def get_application_info(self):
        #~ return (__name__,__version__,__url__)
      
    def get_main_action(self,user):
        return self.modules.lino.Home.default_action
        
    #~ def setup_quicklinks(self,ui,user,tb):
        #~ tb.add_action(self.modules.contacts.Persons.detail_action)
        
    def setup_choicelists(self):
        """
        Defines application-specific default user profiles.
        Local site administrators can override this in their :xfile:.
        """
        from lino import dd
        from django.utils.translation import ugettext_lazy as _
        dd.UserProfiles.reset('* office accounting')
        add = dd.UserProfiles.add_item
        add('000', _("Anonymous"),       '_ _ _', 'anonymous',readonly=True,authenticated=False)
        add('100', _("User"),            'U U U', 'user')
        add('900', _("Administrator"),   'A A A', 'admin')
        
            
    def get_installed_apps(self):
        for a in super(Lino,self).get_installed_apps():
            yield a
        yield 'django.contrib.contenttypes'
        yield 'lino.modlib.users'
        yield 'lino.modlib.countries'
        #~ yield 'lino.modlib.properties'
        yield 'lino.modlib.contacts'
        #~ yield 'lino.modlib.households'
        yield 'lino.modlib.products'
        yield 'lino.modlib.accounts'
        yield 'lino.modlib.ledger'
        yield 'lino.modlib.vat'
        #~ 'lino.modlib.journals',
        yield 'lino.modlib.sales'
        yield 'lino.modlib.finan'
        #~ 'lino.modlib.projects',
        #~ yield 'lino.modlib.blogs'
        #~ yield 'lino.modlib.tickets'
        #~ 'lino.modlib.links',
        #~ yield 'lino.modlib.uploads'
        #~ 'lino.modlib.thirds',
        #~ yield 'lino.modlib.cal'
        #~ yield 'lino.modlib.outbox'
        #~ yield 'lino.modlib.postings'
        yield 'lino.modlib.pages'
        yield 'lino.apps.cosi'



# -*- coding: UTF-8 -*-
## Copyright 2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.


from lino.utils.instantiator import Instantiator, i2d
from lino.core.modeltools import resolve_model
from django.utils.translation import ugettext_lazy as _


from django.conf import settings
from lino.utils.babel import babel_values, babelitem


def objects():
  
    mailType = Instantiator('notes.NoteType').build
    
    yield mailType(**babel_values('name',
        en="Enrolment",
        fr=u'Inscription',de=u"Einschreibeformular"))
    yield mailType(**babel_values('name',
        en="Timetable",
        fr=u'Horaire',de=u"Stundenplan"))
    yield mailType(**babel_values('name',
        en="Letter",
        fr=u'Lettre',de=u"Brief"))



# -*- coding: UTF-8 -*-
## Copyright 2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

raise Exception("No longer user after 20120217")

import logging
logger = logging.getLogger(__name__)

#~ from django.db import models
#~ from django.conf import settings
#~ from django.utils.translation import ugettext as _
#~ from django.utils.encoding import force_unicode 
#~ from django.core.management import call_command

from django.core.management.base import BaseCommand, CommandError
#~ from django.db.models import loading
#~ from django.db import models
#~ from django.contrib.contenttypes.models import ContentType
#~ from lino.core import actors
#~ from lino.utils import get_class_attr

import lino

class Command(BaseCommand):
    help = """Generates a .py file for each .dtl file.
    """
    requires_model_validation = False
    
    def handle(self, *args, **options):
        if len(args) != 0:
            raise CommandError("This command takes no arguments.")
            
        logger.info("Running %s.", self)
        from lino.ui.extjs3 import UI
        #~ UI = settings.LINO.get_ui_class
        ui = UI(make_messages=True)
        #~ # install Lino urls under root location (`/`)
        #~ ui = urlpatterns = ui.get_patterns()
        #~ settings.LINO.setup()
        ui.make_dtl_messages()
        ui.make_linolib_messages()
        
        logger.info("%s done.", self)



# -*- coding: UTF-8 -*-
## Copyright 2008-2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

"""

"""
import datetime

from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from lino import dd
#~ from lino.utils.mti import EnableChild
from lino.modlib.countries.models import CountryCity
from lino.utils import babel 
from lino.models import get_site_config
from lino.utils.choicelists import Gender
from lino.utils import join_words




#~ class AddressType(babel.BabelNamed):
    #~ class Meta:
        #~ verbose_name = _("Address Type")
        #~ verbose_name_plural = _("Address Types")
   
#~ class AddressTypes(dd.Table):
    #~ model = AddressType
    #~ column_names = 'name *'
    
    
class Role(babel.BabelNamed):
    class Meta:
        verbose_name = _("Role")
        verbose_name_plural = _("Roles")
class Roles(dd.Table):
    model = Role
    column_names = 'name *'


class Address(CountryCity):
  
    class Meta:
        abstract = settings.LINO.abstract_address
        verbose_name = _("Address")
        verbose_name_plural = _("Addresses")
  
    addr1 = models.CharField(_("Address line before street"),
        max_length=200,blank=True)
    
    street_prefix = models.CharField(_("Street prefix"),max_length=200,blank=True,
        help_text="Text to print before name of street, but to ignore for sorting.")
    
    street = models.CharField(_("Street"),max_length=200,blank=True,
        help_text="Name of street. Without house number.")
    
    street_no = models.CharField(_("No."),max_length=10,blank=True,
        help_text="House number")
    
    street_box = models.CharField(_("Box"),max_length=10,blank=True,
        help_text="Text to print after :attr:`steet_no` on the same line")
    
    addr2 = models.CharField(_("Address line after street"),
        max_length=200,blank=True,
        help_text="Address line to print below street line")
    
    zip_code = models.CharField(_("Zip code"),
        max_length=10,blank=True)
    region = models.CharField(_("Region"),
        max_length=200,blank=True)
    
    def __unicode__(self):
        return self.address_location(', ')

    def address_person_lines(self):
        #~ yield self.name
        yield self.get_full_name()
        
    def get_full_name(self,*args,**kw):
        return self.name
    full_name = property(get_full_name)
        
        
    def address_location_lines(self):
        if self.addr1:
            yield self.addr1
        if self.street:
            yield join_words(
              self.street_prefix, self.street,
              self.street_no,self.street_box)
        if self.addr2:
            yield self.addr2
        if self.region: # format used in Estonia
            if self.city:
                yield unicode(self.city)
            s = join_words(self.zip_code,self.region)
        else: 
            s = join_words(self.zip_code,self.city)
        if s:
            yield s 
        if self.country is not None:
            sc = get_site_config()
            if not sc.site_company or self.country != sc.site_company.address.country: 
                yield unicode(self.country)
            
        
    def address_lines(self):
        for ln in self.address_person_lines() : yield ln
        for ln in self.address_location_lines() : yield ln
          
    def postal_address(self,linesep="\n"):
        """
        The plain text full postal address (person and location). 
        Lines are separated by `linesep`.
        """
        return linesep.join(list(self.address_person_lines()) + list(self.address_location_lines()))
    postal_address.return_type = models.TextField(_("Address"))
    
    def address_location(self,linesep="\n"):
        """
        The plain text postal address location part. 
        Lines are separated by `linesep`.
        """
        return linesep.join(self.address_location_lines())
        
class AddressDetail(dd.FormLayout):
    address = """
    country region city zip_code
    addr1
    street_prefix street street_no street_box
    addr2
    """
    
if settings.LINO.abstract_address:
  
    AddressableMixin = Address
    
else:  
  
    class AddressableMixin(dd.Model):
        class Meta:
            abstract = True
        address = models.ForeignKey(Address,null=True,blank=True)
        
    class Addresses(dd.Table):
        model = Address
        detail_layout = AddressDetail(main="address")
    
    


class Person(AddressableMixin):
    class Meta:
        verbose_name = _("Person")
        verbose_name_plural = _("Persons")
        
    title = models.CharField(_('Title'),max_length=200,blank=True,
        help_text="""
        Text to print as part of the first address line in front of first_name.
        """)
    first_name = models.CharField(max_length=100)
    last_name = models.CharField(max_length=100)
    gender = Gender.field()
    def __unicode__(self):
        return self.first_name + ' ' + self.last_name
        
        
    
class PersonDetail(AddressDetail):
    main = """
    title first_name last_name gender
    address
    ContactsByPerson
    """
    
  
class Persons(dd.Table):
    model = Person
    column_names = 'last_name first_name *'
    detail_layout = PersonDetail()


class Company(AddressableMixin):  
    class Meta:
        verbose_name = _("Company")
        verbose_name_plural = _("Companies")
        
    name = models.CharField(max_length=50)
        
    def __unicode__(self):
        return self.name
        
class CompanyDetail(AddressDetail):
    main = """
    name
    address
    ContactsByCompany
    """
    
class Companies(dd.Table):
    model = Company
    column_names = 'name *'
    detail_layout = CompanyDetail()




class Contact(dd.Model):  
    class Meta:
        #~ abstract = True
        verbose_name = _("Contact")
        verbose_name_plural = _("Contacts")
        
    person = models.ForeignKey(Person,null=True,blank=True)
    company = models.ForeignKey(Company,null=True,blank=True)
    role = models.ForeignKey(Role,null=True,blank=True)
    
    language = babel.LanguageField()
    
    #~ address = models.ForeignKey(Address,null=True,blank=True)
    #~ address_type = models.ForeignKey(AddressType,blank=True,null=True)
    
    email = models.EmailField(_('E-Mail'),blank=True) # ,null=True)
    url = models.URLField(_('URL'),blank=True)
    phone = models.CharField(_('Phone'),max_length=200,blank=True)
    gsm = models.CharField(_('GSM'),max_length=200,blank=True)
    fax = models.CharField(_('Fax'),max_length=200,blank=True)
    
    def send_mail(self, subject, message, from_email=None):
        "Sends an e-mail to this User."
        from django.core.mail import send_mail
        send_mail(subject, message, from_email, [self.email])
        

    
class Contacts(dd.Table):
    model = Contact
    #~ column_names = '*'
    detail_layout = """
    person company role
    email url
    phone gsm fax
    """

class ContactsByPerson(Contacts):
    master_key = 'person'
    column_names = "email url phone gsm fax company role *"
    
class ContactsByCompany(Contacts):
    master_key = 'company'
    column_names = "email url phone gsm fax person role *"


#~ class User(Person):
class User(Contact):
  
    preferred_foreignkey_width = 15 
    
    class Meta:
        verbose_name = _('User')
        verbose_name_plural = _('Users')

    username = models.CharField(_('username'), max_length=30, 
        unique=True, 
        help_text=_("""
        Required. 30 characters or fewer. 
        Letters, numbers and @/./+/-/_ characters
        """))
    is_staff = models.BooleanField(_('is staff'), default=False, 
        help_text=_("""
        Designates whether the user can log into this admin site.
        """))
    is_expert = models.BooleanField(_('is expert'), default=False, 
        help_text=_("""
        Designates whether this user has access to functions that require expert rights.
        """))
    is_active = models.BooleanField(_('is active'), default=True, 
        help_text=_("""
        Designates whether this user should be treated as active. 
        Unselect this instead of deleting accounts.
        """))
    is_superuser = models.BooleanField(_('is superuser'), 
        default=False, 
        help_text=_("""
        Designates that this user has all permissions without 
        explicitly assigning them.
        """))
    last_login = models.DateTimeField(_('last login'), default=datetime.datetime.now)
    date_joined = models.DateTimeField(_('date joined'), default=datetime.datetime.now)

    def __unicode__(self):
        return self.username

    def full_clean(self,*args,**kw):
        if not self.username and self.person:
            self.username = self.person.first_name.lower()
        super(User,self).full_clean(*args,**kw)
        

class Users(Contacts):
    model = User
    column_names = 'username person company is_superuser is_staff *'

    
if settings.LINO.is_installed('contacts'):
    """
    Don't inject fields if contacts is just being imported from some other module.
    """
  
    from lino.models import SiteConfig

    dd.inject_field(SiteConfig,
        'site_company',
        models.ForeignKey(Company,
            blank=True,null=True,
            verbose_name=_("The company that runs this site"),
            related_name='site_company_sites',
            ),
        """The Company to be used as sender in documents.""")
        
    
        
        
NAME = "contacts"
LABEL = _("Contacts")

def setup_main_menu(site,ui,profile,m):
    m.add_action(Companies)
    m.add_action(Persons)
    #~ m.add_action(Contacts)
    

def setup_my_menu(site,ui,profile,m): 
    pass
  
def setup_config_menu(site,ui,profile,m): 
    #~ m = m.add_menu(NAME,LABEL)
    #~ m.add_action(AddressTypes)
    m.add_action(Roles)
    m.add_action(site.modules.countries.Countries)
    m.add_action(site.modules.countries.Cities)
            
  
def setup_explorer_menu(site,ui,profile,m):
    #~ m = m.add_menu(NAME,LABEL)
    m.add_action(Contacts)
    if not settings.LINO.abstract_address:
        m.add_action(Addresses)
  


## Copyright 2009-2012 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

import os
from django.conf import settings
from lino.ui import base
#~ from lino.ui.console import renderers_text

#~ class Panel:
    #~ pass
    
#~ class Element:
    #~ def __init__(self,*args,**kw):
        #~ self.args = args
        #~ self.kw = kw

#~ class FieldElement(Element):
    #~ def __init__(self,field,*args,**kw):
        #~ self.field = field
        #~ Element.__init__(self,*args,**kw)

#~ class MethodElement(Element):
    #~ def __init__(self,field,*args,**kw):
        #~ self.field = field
        #~ Element.__init__(self,*args,**kw)

#~ class ConsoleUI(base.UI):
  
    #~ GridElement = Element
    #~ VirtualFieldElement = Element
    #~ MethodElement = Element
    #~ ButtonElement = Element
    #~ Panel = Panel
    #~ Store = Element
    #~ Spacer = Element
    #~ StaticTextElement = Element
    #~ InputElement = Element
  
    #~ def main_panel_class(self,layout):
        #~ return Panel
    #~ def report_as_text(self,rpt):
        #~ rh = self.get_report_handle(rpt)
        #~ rr = renderers_text.TextReportRequest(rh,*args,**kw)
        #~ return rr.render()
        
from lino.utils.tables import TableRequest
#~ from lino.extjs import ui as extui

class PseudoRequest:
    def __init__(self,name):
        self.user = settings.LINO.user_model.objects.get(username=name)


   
class Console(base.UI):

#~ class Console(object):
    _handle_attr_name = '_console_handle'
    
    def __init__(self):
        #~ settings.LINO.setup()
        #~ self.request = PseudoRequest(username)
        base.UI.__init__(self)
        self.create_layout_element = settings.LINO.ui.create_layout_element
        self.create_layout_panel = settings.LINO.ui.create_layout_panel
        
    def setup_handle(self,h,ar):
        settings.LINO.ui.setup_handle(h,ar)
        
    #~ def create_layout_element(self,lh,name,**kw):
        #~ return extui.create_layout_element(lh,name,**kw)
        
    #~ def request(self,actor,**kw):
        #~ if isinstance(actor,basestring):
            #~ actor = settings.LINO.modules.resolve(actor)
        #~ return actor.request(self,**kw)
        
    def run(self,action,pk):
        elem = action.actor.model.objects.get(pk=pk)
        #~ ar = TableRequest(self,action.actor,self.request,action)
        ar = TableRequest(self,action.actor,None,action)
        return action.run(elem,ar)
        
    def action_response(self,kw):
        """
        checking first whether there are only allowed keys 
        (defined in :attr:`ACTION_RESPONSES`)
        """
        self.check_action_response(kw)
        msg = kw.get('message')
        if msg: 
            print msg
        url = kw.get('open_url') or kw.get('open_davlink_url')
        if url:
            os.startfile(url)
        return kw
        

#~ ui = ConsoleUI()


## Copyright 2009-2011 Luc Saffre
## This file is part of the Lino project.
## Lino is free software; you can redistribute it and/or modify 
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 3 of the License, or
## (at your option) any later version.
## Lino is distributed in the hope that it will be useful, 
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
## GNU General Public License for more details.
## You should have received a copy of the GNU General Public License
## along with Lino; if not, see <http://www.gnu.org/licenses/>.

import logging
logger = logging.getLogger(__name__)

import os
import sys
from django.conf import settings
from django.conf.urls.defaults import *
from django.contrib import databrowse
#~ from django.contrib.auth.decorators import login_required
#~ from django.contrib.auth import urls as auth_urls
from django.utils import importlib
import lino
from . import ext_ui 

urlpatterns = patterns('',
    (r'^favicon\.ico$', 'django.views.generic.simple.redirect_to', 
        {'url': settings.MEDIA_URL + 'lino/favicon.ico'})
)

    

settings.LINO.setup()

ui = ext_ui.ExtUI(settings.LINO)

urlpatterns += patterns('',
    (r'', include(ui.get_urls())),
)

if sys.platform == 'win32':

    EXTJS_ROOT = r's:\ext-4.0.0'
    #~ EXTJS_ROOT = r's:\ext-3.3.1'
    #~ EXTJS_ROOT = r's:\ext-3.3.0'
    #~ EXTJS_ROOT = r's:\ext-3.2.1'
    #~ EXTJS_ROOT = r's:\ext-3.3.0-rc'
    #~ EXTJS_URL = "/media/extjs/"

    LINO_MEDIA = os.path.abspath(os.path.join(
        os.path.dirname(lino.__file__),'..','media'))
    #~ LINO_MEDIA = os.path.join(lino_site.lino_site.ui.source_dir(),'media')
    #~ print 'LINO_MEDIA=',LINO_MEDIA
    
    if not os.path.exists(EXTJS_ROOT):
        raise Exception("EXTJS_ROOT %s does not exist" % EXTJS_ROOT)
        #~ logger.warning("EXTJS_ROOT %s does not exist",EXTJS_ROOT)
        
    prefix = settings.MEDIA_URL[1:]
    assert prefix.endswith('/')
    
    urlpatterns += patterns('django.views.static',
    (r'^%sextjs/(?P<path>.*)$' % prefix, 
        'serve', {
        'document_root': EXTJS_ROOT,
        'show_indexes': True }),)
        
    urlpatterns += patterns('django.views.static',
    (r'^%slino/(?P<path>.*)$' % prefix, 
        'serve', {
        'document_root': LINO_MEDIA,
        'show_indexes': True }),)

    #~ print LINO_MEDIA
  
    urlpatterns += patterns('django.views.static',
        (r'^%s(?P<path>.*)$' % prefix, 'serve', { 
          'document_root': settings.MEDIA_ROOT, 
          'show_indexes': True }),
        #~ (r'^lino_media/(?P<path>.*)$', 'serve', { 'document_root': LINO_MEDIA, 'show_indexes': True }),
    )





#~ from lino.apps.min1.settings import *
#~ from lino.apps.min2.settings import *
from lino.apps.presto.settings import *
class Lino(Lino):
    title = "mini tutorial"
    languages = ['en', 'de','fr']
    #~ languages = ['en']
    
LINO = Lino(__file__,globals()) 

DEBUG = True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3', 
        'NAME': 'test.db',
    }
}

LOGGING = dict(level='DEBUG')
#~ LOGGING = dict(filename=filename,level='DEBUG',rotate=False)  
#~ LOGGING = dict(filename=join(LINO.project_dir,'log',filename),level='DEBUG')
#~ LOGGING = dict(filename=join(LINO.project_dir,'log',filename),level='DEBUG')
import datetime
filename = datetime.date.today().strftime('%Y-%m-%d.log')
logdir = join(LINO.project_dir,'log')
import os
if os.path.exists(logdir):
    LOGGING.update(filename=join(logdir,filename))


EMAIL_HOST = "your.smtp.host"
SERVER_EMAIL = 'you@example.com'

# uncomment the following line for testing in memory database:
#~ DATABASES['default']['NAME'] = ':memory:'



import os
#~ from xml.dom import minidom
from lxml import etree


XSD_NAMESPACE = "http://www.w3.org/2001/XMLSchema"
XSD = "{%s}" % XSD_NAMESPACE

NSMAP = dict(xsd=XSD_NAMESPACE)


def xsd2py(fn,nsname):
    tree = etree.parse(fn) 
    root = tree.getroot()
    yield "from lino.utils.bcss import Namespace"
    yield "%s = Namespace(%r,%r):" % (nsname,nsname,str(root.get('targetNamespace'))
    for e in root:
        if e.tag == XSD + 'complexType':
            na = e.get('name',None)
            if na is None:
                yield indent + "# skipped nameless %s" % e.tag
                continue
            yield indent + "class %s(xg.Container):" % na
            for ee in e:
                #~ seq = e.getElementsByTagName('xsd:sequence')[0]
                if ee.tag == XSD + 'sequence':
                    #~ seq = e.attributes.get('sequence',None)
                    #~ if seq:
                    for item in ee:
                        na = item.get('name',None)
                        if na is None: continue
                        ta = item.get('type',None)
                        if ta is None:
                            yield indent*2 + "class %s(): pass" % na
                        else:
                            if ':' in ta:
                                typename = ta.replace(':','.')
                            else:
                                typename = nsname + '.' + ta
                            yield indent*2 + "class %s(%s): pass" % (na,typename)
        elif e.tag == XSD + 'simpleType':
            na = e.get('name',None)
            if na is None:
                yield indent + "# skipped nameless %s" % e.tag
                continue
            yield indent + "# %s" % e.tag
            yield indent + "class %s(xg.Container):" % na
        else:
            yield indent + "# unhandled element %s" % e.tag


fn = os.path.join('XSD','RetrieveTIGroupsV3.xsd')
for ln in xsd2py(fn,'tx25'):
    print ln
    